# tested for PhyloNet 3.6.8
from Bio import Phylo
import concurrent, io, itertools, math, os, random, re, shutil, subprocess, sys, time

# call PhyloNet:


def PhyloNet_single_job(trees, PhyloNet_line, step, conf, IO_files):
    """
    Calls PhyloNet for single species tree reconstruction.
    
    Args:
        trees: gene trees in newick format (list of strings)
        PhyloNet_line: instruction for tree inference
        step: name of PhyloNet call
        conf: namespace of program options
        IO_files: set, stores names of all existing PhyloNet input/output
                  files for later cleanup
    
    Side effects:
        PhyloNet output is written, IO_files are updated.
    """
    fname_input = os.path.join(conf.working_directory, step + ".nex")
    fname_output = os.path.join(conf.working_directory, step + ".txt")
    
    with open(fname_input, "w") as f:
        f.write("#NEXUS\n\nBEGIN TREES;\n\n")
        for i, tree in enumerate(trees, 1):
            newick_line = io.StringIO()
            Phylo.write(tree, newick_line, "newick", plain=True)
            f.write("".join(["Tree g", "{0:07d}".format(i), " =\n", newick_line.getvalue()]))
        
        f.write("\nEND;\n\n\nBEGIN PhyloNet;\n\n")
        f.write(PhyloNet_line)
        f.write("\nEND;\n")

    # call PhyloNet
    os.system("".join(["java ", conf.java_options, " -jar ", conf.phylonet_path, 
                       " ", fname_input, " > ", fname_output]))
    
    # update file list
    IO_files.update([fname_input, fname_output])


def PhyloNet_batch(trees, PhyloNet_lines, name, step, input_files, conf, IO_files):
    """
    Calls PhyloNet for multiple species tree reconstructions.
    
    Args:
        trees: gene trees in newick format (list of strings)
        PhyloNet_lines: tree inference instructions
        name: name polyploid taxon
        step: name of PhyloNet calls
        input_files: list, stores names of currently processed 
                     PhyloNet input files
        conf: namespace of program options
        IO_files: set, stores names of all existing PhyloNet input/output
                  files for later cleanup
    
    Side effects:
        PhyloNet output is written, input_files and IO_files are updated.
    """
    input_files.clear()

    fname_head = os.path.join(conf.working_directory, name + "_" + step + "_head.nex")
    
    # write new head if necessary
    if fname_head not in IO_files:
        with open(fname_head, "w") as f:
            f.write("#NEXUS\n\nBEGIN TREES;\n\n")
            # write trees
            for i, tree in enumerate(trees, 1):
                newick_line = io.StringIO()
                Phylo.write(tree, newick_line, "newick", plain=True)
                f.writelines(["Tree g", "{0:07d}".format(i), " =\n", newick_line.getvalue()])
            f.write("\nEND;\n\n\n\nBEGIN PhyloNet;\n\n")
    
    chunk_size = max(len(PhyloNet_lines) // conf.max_procs, 1)
    n_chunks = math.ceil(len(PhyloNet_lines) / chunk_size)

    # split PhyloNet input into chunks
    for i in range(n_chunks):
        fname_chunk = os.path.join(conf.working_directory, name + "_" + step + "_stack{0:04d}.nex".format(i))
        shutil.copyfile(fname_head, fname_chunk)  # copy trees

        with open(fname_chunk, "a") as f:
            f.writelines(PhyloNet_lines[i*chunk_size : (i+1)*chunk_size])  # (out of range indices will be omitted)
            f.write("\n\nEND;\n")
        input_files.append(fname_chunk)

    processes = set()

    for infile in input_files:
        outfile = os.path.splitext(infile)[0] + ".txt"

        with open(outfile, "w") as fout:
            processes.add(subprocess.Popen(
                ["java", conf.java_options, "-jar", conf.phylonet_path, infile], stdout=fout
            ))

        if len(processes) >= conf.max_procs:
            os.wait()
            processes.difference_update(
                [p for p in processes if p.poll() is not None])

    # Check if all the child processes were closed
    for p in processes:
        if p.poll() is None:
            p.wait()
    
    # update file list
    IO_files.add(fname_head)
    IO_files.update(input_files)
    IO_files.update([os.path.splitext(fname)[0] + ".txt" for fname in input_files])


# get results:

def parse_output_chunk(data_chunk):
    """
    Parses PhyloNet output for one species tree reconstruction.
    
    Args:
        data chunk: list of four consecutive lines 
                    from PhyloNet output file
    
    Returns:
        species tree, input tree IDs, allele mapping, 
        number of extra lineages
    """
    match_tree = re.search("^(\\(.+;)\n?", data_chunk[2])
    match_ids = re.search("\\(([ ,g0-9]*)\\)", data_chunk[1])
    match_map = re.search("<(.+)>", data_chunk[1])
    match_dist = re.search("Total number of extra lineages:(.+)\n?", data_chunk[3])

    # check for valid matches and additional empty line
    if match_tree and match_ids and match_map and match_dist and data_chunk[0]=="\n":
        chunk_tree = match_tree.group(1)
        chunk_ids = match_ids.group(1)
        chunk_map = match_map.group(1)
        chunk_dist = int(float(match_dist.group(1)))
    else:
        sys.exit("error: cannot parse PhyloNet output!")
    
    return chunk_tree, chunk_ids, chunk_map, chunk_dist


def get_results_single(step, conf):
    """
    Evaluates PhyloNet output (single species tree reconstruction).
    
    Args:
        step: name of PhyloNet call
        conf: namespace of program options
    
    Returns:
        species tree, number of extra lineages, allele mapping
    """
    fname = os.path.join(conf.working_directory, step + ".txt")

    with open(fname, "r") as f:
        data_chunk = f.readlines()

    if not data_chunk:
        sys.exit("error: cannot parse PhyloNet output!")
    
    tree, ids, mapping, dist = parse_output_chunk(data_chunk)
    
    return tree, dist, mapping


def get_results_multi(input_files, n_PhyloNet_instructions, conf):
    """
    Evaluates PhyloNet output (multiple species tree reconstructions).
    
    Args:
        input_files: names of input files, which were passed to PhyloNet
        n_PhyloNet_instructions: number of species tree reconstructions
        conf: namespace of program options
    
    Returns:
        tuple of three equal-length lists:
        species trees, numbers of extra lineages, allele mappings 
    """
    chunk_size = max(1, n_PhyloNet_instructions // conf.max_procs)
    
    distances = [0] * n_PhyloNet_instructions
    trees = [0] * n_PhyloNet_instructions
    mappings = [0] * n_PhyloNet_instructions
    
    i = 0
    for fname in input_files:
        with open(os.path.splitext(fname)[0] + ".txt", "r") as f:
            for j in range(chunk_size):
                # read 4 lines from file
                data_chunk = [line for line in itertools.islice(f, 4)]
                
                if not data_chunk:
                    if (fname != input_files[-1]) or (j != n_PhyloNet_instructions % chunk_size):
                        sys.exit("error: cannot parse PhyloNet output!")
                    break
                
                trees[i], _, mappings[i], distances[i] = parse_output_chunk(data_chunk)
                i += 1
    return trees, distances, mappings




#!/usr/bin/env python3

import argparse

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("tree_file", help="Species trees with inferred pseudo-diploids, one newick string per line.")
    parser.add_argument("taxon", help="Name of the taxon which has been split into pseudo-diploids.")
    parser.add_argument("permutation_file", help="Alignment of clusters among multiple clusterings.")
    args = parser.parse_args()

    with open(args.tree_file, "r") as treefile:
        with open(args.permutation_file, "r") as permfile:
            for tree, perm in zip(treefile, permfile):
                if not (tree and perm):
                    break
                
                perm = perm.strip().split("\t")
                tree_relabeled = tree
                
                for i, cluster in enumerate(perm):
                    old = args.taxon  + "___" + "{:02}".format(int(cluster))
                    new = args.taxon  + "_P" + str(i+1)
                    tree_relabeled = tree_relabeled.replace(old, new)
                
                print(tree_relabeled, end="")

                
if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import configargparse, numpy, random, time
import collections, copy, itertools, os, re, sys
from Bio import Phylo
from allcopol.utils import *


def main():

    cparser = configargparse.ArgParser()

    cparser.add_argument(
        "-C", "--config-file", 
        required = False, 
        is_config_file = True
    )
    
    # obligate parameters:
    
    cparser.add_argument(
        "-A", "--allele-mapping", 
        required = True
    )
    cparser.add_argument(
        "-G", "--gene-trees", 
        required = True, 
        help = "Gene tree file (1 newick line per tree)."
    )
    cparser.add_argument(
        "-S", "--tree-sample-size", 
        required = True, 
        type = int, 
        help = "Number of gene trees per locus."
    )
    cparser.add_argument(
        "-P", "--phylonet-path", 
        required = True, 
        help = "Path to PhyloNet jar file."
    )

    # optional parameters:
    
    cparser.add_argument(
        "-o", "--outgroup", 
        required = False, 
        nargs = '+', 
        default = [],
        help = "Root species trees at outgroup. Warning: If multiple taxa are specified, \
        it is not checked, whether non-outgroup taxa are monophyletic."
    )
    cparser.add_argument(
        "-p", "--polyploids", 
        required = False, 
        nargs = '+', 
        default = [],
        help = "If specified, only explicitly mentioned polyploid taxa will be analyzed \
        (if present in the mapping file). The option can be used to entirely skip the \
        optimization procedure (eg. \"-p none\")"
    )
    cparser.add_argument(
        "-m", "--max-procs", 
        required = False, 
        type = int,
        default = 1,
        help = "Maximum number of PhyloNet instances to be run in parallel. \
        Warning: (Over)parallelization causes significant computational overhead."
    )
    cparser.add_argument(
        "-t", "--tabu-tenure", 
        required = False, 
        type = int, 
        default = 2,
        help = "Length of the tabu list (iterations)."
    )
    cparser.add_argument(
        "-s", "--nb-sample-size", 
        required = False, 
        type = float, 
        default = "inf",
        help = "(Max.) number of evaluated candidate solutions per iteration. Default: \"inf\" \
        (complete neighborhood)"
    )
    cparser.add_argument(
        "-i", "--max-iter", 
        required = False, 
        type = int, 
        default = 500,
        help = "Maximum number of tabu search iterations."
    )
    cparser.add_argument(
        "-u", "--max-unimproved", 
        required = False, 
        type = float, 
        default = "inf",
        help = "Reinitialize search if no improvement has been found for MAX_UNIMPROVED iterations."
    )
    cparser.add_argument(
        "-w", "--working-directory", 
        required = False, 
        default = ""
    )
    cparser.add_argument(
        "-c", "--phylonet-command", 
        required = False, 
        default = "Infer_ST_MDC", 
        help = "Only tested with Infer_ST_MDC (default)!"
    )
    cparser.add_argument(
        "-f", "--remove-files", 
        required = False, 
        action = "store_true", 
        default = False, 
        help = "When finished, remove every file in WORKING_DIRECTORY except results.txt"
    )
    cparser.add_argument(
        "-j", "--java-options", 
        required = False, 
        default = "-Xmx1200m"
    )
    cparser.add_argument(
        "-b", "--display-branch-length", 
        required = False, 
        action = "store_true", 
        default = False
    )
    cparser.add_argument(
        "-a", "--print-ascii-tree", 
        required = False, 
        action = "store_true", 
        default = False
    )
    cparser.add_argument(
        "-r", "--report-interval", 
        required = False, 
        type = float, 
        default = "inf", 
        help = "Print intermediate results every report_interval iterations."
    )
    cparser.add_argument(
        "-e", "--evaluate-initial", 
        required = False, 
        action = "store_true", 
        default = False, 
        help = "Evaluate initial solution (iteration 0) and report results."
    )
    cparser.add_argument(
        "-d", "--seed", 
        required = False, 
        type = int, 
        default = -1, 
        help = "Initialize pseudorandom number generator, default: -1 (disabled)"
    )
    cparser.add_argument(
        "--diploids-only", 
        action = "store_true", 
        required = False, 
        default = False, 
        help = "Calculate species tree without reconstructed pseudo-diploids."
    )
    cparser.add_argument(
        "--combine-taxa", 
        action = "store_true", 
        required = False, 
        default = False, 
        help = "Combine pseudo-diploids inferred for different polyploid taxa into one species tree"
    )

    conf = cparser.parse_args()

    # correct list arguments
    conf.outgroup = [y.strip() for x in conf.outgroup for y in x.split(",") if y.strip()]
    conf.polyploids = [y.strip() for x in conf.polyploids for y in x.split(",") if y.strip()]

    # initialize pseudorandom number generators
    if conf.seed != -1:
        random.seed(conf.seed)
        numpy.random.seed(conf.seed)
    
    # named tuple for evaluated solutions
    evaluated = collections.namedtuple("evaluated", ["solution", "tree", "cost", "iteration"])

    # create working directory
    if not conf.working_directory:
        conf.working_directory = time.strftime("%Y_%h_%d")
    i = 1
    dirname = conf.working_directory
    while os.path.exists(dirname):
        dirname = conf.working_directory + "_" + str(i)
        i += 1
    os.mkdir(dirname)
    conf.working_directory = dirname

    # print/write parsed options
    d = vars(conf)
    
    with open(os.path.join(conf.working_directory, "results.txt"), "w") as f:
        for k in sorted(d.keys()):
            f.write(k + ": " + str(d[k]) + "\n")
            print(k + ": " + str(d[k]))
        f.write("----------------------\n")
        print("----------------------")

    time_start = time.time()

    IO_files = set()

    # read input:
    
    # get name, species membership, allele IDs and ploidy level of each accession
    species_list = read_allele_mapping(conf.allele_mapping)

    species_diploid = list(filter(lambda x: x.ploidy == 2, species_list))
    species_polyploid = list(filter(lambda x: x.ploidy > 2, species_list))

    # create allele mapping string for true diploids (accession-wise)
    mapping_orig = ";".join(acc.name + ":" + ",".join(acc.alleles) for sp in species_diploid for acc in sp.accessions)
    # species-wise
    mapping_orig_spec = ";".join(sp.name + ":" + ",".join(sp.get_alleles()) for sp in species_diploid)

    # read and preprocess gene trees
    input_data_trees = read_trees(conf.gene_trees, species_polyploid, conf.tree_sample_size)
    tree_IDs = ",".join("g" + "{0:07d}".format(i+1) for i in range(len(input_data_trees)))


    if conf.diploids_only:
        print ("\nDiploids only:")
        print("#" * 14, "\n")
        
        # prune gene trees (remove alleles of all polyploid species)
        ptrees = copy.deepcopy(input_data_trees)
        accessions_2b_pruned = [acc for sp in species_polyploid 
                                    for acc in sp.accessions ]
        
        for tree in ptrees:
            for acc in accessions_2b_pruned:
                prune_tree(tree, acc.alleles, "_m[0-9]+")

        # write pruned trees into file (unused)
        fname_ptrees = os.path.join(conf.working_directory,  "diploids_pruned.txt")
        Phylo.write(ptrees, fname_ptrees, "newick", plain=True)
        IO_files.add(fname_ptrees)
        
        PhyloNet_instruction = build_PhyloNet_instruction(tree_IDs, mapping_orig_spec, "", conf)
        # call PhyloNet
        PhyloNet_single_job(ptrees, PhyloNet_instruction, "diploids", conf, IO_files)
        # evaluate output
        stree, dist, _ = get_results_single("diploids", conf)
        print_results("Species tree diploids", stree, mapping_orig_spec, dist, conf)


    for species in species_polyploid:
        if len(conf.polyploids) > 0 and species.name not in conf.polyploids:
            continue
        
        print("\nSpecies:", species.name)
        print("#" * (9+len(species.name)), "\n")
        
        # prune gene trees (remove alleles of all but one polyploid species)
        ptrees = copy.deepcopy(input_data_trees)
        accessions_2b_pruned = [acc for sp in species_polyploid
                                    for acc in sp.accessions
                                    if sp != species]
        
        for tree in ptrees:
            for acc in accessions_2b_pruned:
                prune_tree(tree, acc.alleles, "_m[0-9]+")

        # write pruned trees into file (unused)
        fname_ptrees = os.path.join(conf.working_directory, species.name + "_pruned.txt")
        Phylo.write(ptrees, fname_ptrees, "newick", plain=True)
        IO_files.add(fname_ptrees)
        
        # allele names incl. marker suffix
        alleles_blockwise = []
        
        for accession in species.accessions:
            # loop across loci/markers
            for j in range(0, len(ptrees), conf.tree_sample_size):
                tree = ptrees[j]
                # search patterns matching preprocessed alleles of current accession and marker
                pattern_list = ["^" + re.escape(allele) + "_m[0-9]+$" for allele in accession.alleles]
                alleles_blockwise.append([leaf.name for p in pattern_list 
                                                for leaf in tree.get_terminals()
                                                if re.match(p, leaf.name)])
        
        n_alleles = [len(block) for block in alleles_blockwise]
        
        # initialization
        current_solution = evaluated(initial_solution(species.ploidy, n_alleles), "", float("inf"), 0)
        best_solution = current_solution
        best_solution_run = current_solution
        n_unimproved = 0
        tabu_list = []
        
        if conf.evaluate_initial:
            print("iteration: 0")
            partition = get_allele_partition(current_solution.solution, alleles_blockwise, species.ploidy)
            mapping_hyp = build_mapping(species.name, partition)
            PhyloNet_instruction = build_PhyloNet_instruction(tree_IDs, mapping_orig_spec, mapping_hyp, conf)
            # call PhyloNet
            PhyloNet_single_job(ptrees, PhyloNet_instruction, "initial", conf, IO_files)
            # evaluate output
            stree, dist, _ = get_results_single("initial", conf)
            current_solution = evaluated(current_solution.solution, stree, dist, 0)
            best_solution = current_solution
            
            print("current solution:", str(current_solution.cost), "/ best so far:", str(best_solution.cost))
            print_results(
                "Species tree " + species.name, 
                best_solution.tree, 
                mapping_orig_spec + ";" + mapping_hyp, 
                best_solution.cost, 
                conf
            )
        
        # optimize allele partitions
        for iteration in range(1, conf.max_iter+1):
            
            print("iteration:", str(iteration))
            
            neighborhood = list(neighborhood_feasible(current_solution.solution, species.ploidy))
            ssize = int(min(conf.nb_sample_size, len(neighborhood)))
            candidates, moves = zip(*random.sample(neighborhood, ssize))
            
            partitions = [get_allele_partition(parent_indices, alleles_blockwise, species.ploidy) 
                          for parent_indices in candidates]
            
            PhyloNet_instructions = []
            for p in partitions:
                # allele mapping string for hypothetical parents
                mapping_hyp = build_mapping(species.name, p)
                instr = build_PhyloNet_instruction(tree_IDs, mapping_orig_spec, mapping_hyp, conf)
                PhyloNet_instructions.append(instr)
            
            # call PhyloNet
            input_files = []
            PhyloNet_batch(ptrees, PhyloNet_instructions, species.name, "TS", input_files, conf, IO_files)
            
            # evaluate output
            strees, dist, _ = get_results_multi(input_files, len(candidates), conf)

            updated = False
            n_rejected = 0

            for j in numpy.argsort(dist):
                if not is_tabu(moves[j], tabu_list) or dist[j] < best_solution_run.cost:  # best_solution.cost:
                    current_solution = evaluated(candidates[j], strees[j], dist[j], iteration)
                    tabu_list = ([[reverse_move(m) for m in moves[j]]] + tabu_list) [:conf.tabu_tenure]
                    updated = True
                    
                    if current_solution.cost < best_solution_run.cost:
                        best_solution_run = current_solution
                        n_unimproved = 0
                    else:
                        n_unimproved += 1
                        
                    if current_solution.cost < best_solution.cost:
                        best_solution = current_solution
                        
                    break
                else:
                    n_rejected += 1
            
            if not updated:
                tabu_list = tabu_list[:-1]
            
            species.partition = get_allele_partition(best_solution.solution, alleles_blockwise, species.ploidy)
            if n_rejected > 0:
                print("    " + str(n_rejected) + " move(s) rejected")
            print("    current solution:", str(current_solution.cost), "/ best so far:", str(best_solution.cost))
            
            # reinitialization
            if n_unimproved >= conf.max_unimproved:
                print("\n---\nreinitialization\n---\n")
                current_solution = evaluated(initial_solution(species.ploidy, n_alleles), "", float("inf"), 0)
                best_solution_run = current_solution
                n_unimproved = 0
                tabu_list = []
            
            # report intermediate results
            if iteration % conf.report_interval == 0 and iteration < conf.max_iter:
                mapping_hyp = build_mapping(species.name, species.partition)
                print_results(
                    "Species tree " + species.name, 
                    best_solution.tree, 
                    mapping_orig_spec + ";" + mapping_hyp, 
                    best_solution.cost, 
                    conf
                )
            
        # report results for current polyploid
        print("\n---------------\n\nBest solution found in iteration", str(best_solution.iteration))
        
        mapping_hyp = build_mapping(species.name, species.partition)
        print_results(
            "Species tree " + species.name, 
            best_solution.tree, 
            mapping_orig_spec + ";" + mapping_hyp, 
            best_solution.cost, 
            conf
        )

    # species tree for multiple polyploids
    if conf.combine_taxa:
        print ("\nCombined polyploids:")
        print("#" * 20, "\n")
        
        # prune gene trees (remove alleles of all but one polyploid species)
        ptrees = copy.deepcopy(input_data_trees)
        accessions_2b_pruned = [acc for sp in species_polyploid 
                                    for acc in sp.accessions 
                                    if len(conf.polyploids) > 0 and sp.name not in conf.polyploids]
        
        for tree in ptrees:
            for acc in accessions_2b_pruned:
                prune_tree(tree, acc.alleles, "_m[0-9]+")

        # allele mapping string for hypothetical ancestors
        mapping_hyp = []
        for species in species_polyploid:
            if species.name in conf.polyploids or len(conf.polyploids) == 0:
                mapping_hyp.append(build_mapping(species.name, species.partition))
        mapping_hyp = ";".join(mapping_hyp)

        PhyloNet_instruction = build_PhyloNet_instruction(tree_IDs, mapping_orig_spec, mapping_hyp, conf)

        # call PhyloNet
        PhyloNet_single_job(ptrees, PhyloNet_instruction, "combined", conf, IO_files)

        # evaluate output
        stree, dist, _ = get_results_single("combined", conf)
        print_results(
            "Species tree combined", 
            stree,
            mapping_orig_spec + ";" + mapping_hyp, 
            dist, 
            conf
        )

    # final cleanup
    if conf.remove_files:
        for fname in IO_files:
            os.remove(fname)
                    
    time_stop = time.time()
    
    print("Total Time elapsed:", "%.2f" % (time_stop-time_start), "s")

if __name__ == "__main__":
    main()



import collections, itertools, random


def initial_solution(ploidy, n_alleles):
    """
    Creates random feasible solution.
    
    Args:
        ploidy: determines the number of pseudo-diploids (n/2)
        n_alleles: list, specifies how many alleles have to be 
                   partitioned (one entry per accession and marker)
    Returns:
        pseudo-diploid indices for each allele as list of lists
        
    Example:
        >>> initial_solution(4, [4,4])
        [[0,1,1,0], [1,1,0,0]]
    """
    start = []
    
    for n in n_alleles:
        parents = [i for i in range(ploidy//2) for j in range(2)]
        parents = random.sample(parents, n)
        start.append(parents)
    
    return start


def neighborhood_feasible(solution, ploidy):
    """
    Generates all feasible solutions in the neighborhood of an existing solution.
    
    Args:
        solution: current feasible solution
        ploidy
        
    Yields:
        new candidate solution along with 
        corresponding moves (locus, allele, FROM, TO)
    
    Example:
        >>> nb = neighborhood_feasible([[0,1,1,0], [1,1,0,0]], 4)   
        >>> nb.__next__()  # (new_solution, (move1, move2))
        ([[1, 0, 1, 0], [1, 1, 0, 0]], ((0, 0, 0, 1), (0, 1, 1, 0)))
    """
    # shift parent (if possible)
    for i, locus in enumerate(solution):
        c = collections.Counter(locus)
    
        # find accessible parents (i.e. less than 2 alleles)
        for j in range(ploidy//2):
            if c[j] < 2:
                # shift alleles from other parents
                for k, parent in enumerate(locus):
                    if parent != j:
                        locus_new = solution[i][:k] + [j] + solution[i][k+1:]
                        solution_new = solution[:i] + [locus_new] + solution[i+1:]
                        moves = ((i, k, parent, j),)  # locus, allele, FROM, TO
                        yield(solution_new, moves)
                
    # swap parents
    for i, locus in enumerate(solution):
        # group alleles by parent
        allele_groups = [[j for j in range(len(locus)) if locus[j] == parent] 
                            for parent in set(locus)]
        for comb in itertools.combinations(allele_groups, 2):
            # build pairs of alleles from different parents
            for pair in itertools.product(*comb):
                i0 = pair[0]
                i1 = pair[1]
                locus_new = locus[:]
                locus_new[i0], locus_new[i1] = locus_new[i1], locus_new[i0]
                solution_new = solution[:i] + [locus_new] + solution[i+1:]
                # (locus, allele, FROM, TO)
                moves = ((i, i0, solution[i][i0], solution[i][i1]), 
                         (i, i1, solution[i][i1], solution[i][i0]))
                yield(solution_new, moves)


def reverse_move(move):
    """
    Swaps FROM and TO from move.
    
    Args:
        move: tuple of indices (locus, allele, FROM, TO)
        
    Returns:
        reverse move
    """
    return move[0], move[1], move[3], move[2]


def is_tabu(moves, tabu_list):
    """
    Checks if moves contain any tabu-active allele-parent assignment.
    
    Args:
        moves: list [(locus, allele, FROM, TO), ...]
        
    Returns:
        tabu status
    """
    forbidden_assignments = [(locus, allele, TO) for (locus, allele, FROM, TO) in itertools.chain(*tabu_list)]
    tabus = [(m[0], m[1], m[3]) in forbidden_assignments for m in moves]
    
    if any(tabus):
        return True
    else:
        return False


def get_allele_partition(parent_indices, alleles_blockwise, ploidy):
    """
    Concatenates alleles names for each pseudo-diploid into strings.
    
    Args:
        parent_indices: pseudo-diploid indices for each allele 
                        as list of lists
        alleles_blockwise: suffixed allele names as list of lists,
                           one entry per accession and marker
        ploidy
    
    Returns:
        list of allele mapping strings for each pseudo-diploid
    
    Example:
        >>> get_allele_partition(
        ...     [[0,1,1], [0,0]],
        ...     [["a1_m0", "a2_m0", "a3_m0"], ["a1_m1", "a2_m1"]],
        ...     4
        ... )
        ['a1_m0,a1_m1,a2_m1', 'a2_m0,a3_m0']
    """
    partition = [[] for i in range(ploidy//2)]
    
    for i, block in enumerate(alleles_blockwise):
        for j, allele in enumerate(block):
            partition[parent_indices[i][j]].append(allele)
    
    # one string per parent
    partition = [",".join(parent) for parent in partition]
    return partition




import io, os, re, sys
from Bio import Phylo

# classes for storage of accession/taxon-specific information


class Accession:
    def __init__(self, name, alleles):
        self.name = name
        self.alleles = alleles
        self.partition = []


class Species:
    def __init__(self, species, ploidy, accession, alleles):
        self.name = species
        self.ploidy = ploidy
        self.accessions = [Accession(accession, alleles)]
        self.partition = []

    def get_alleles(self):
        """
        Summarizes allele IDs from all accessions of the species.
        """
        return sorted(set(sum((acc.alleles for acc in self.accessions), [])))


def clearStringIO(s):
    s.seek(0)
    s.truncate(0)


def rename_polyploids(input_data, species_polyploid, tree_repeats):
    """
    Appends locus/marker index (e.g. "_m2") to allele IDs from polyploids.
    
    Args:
        input_data: list of newick strings
        species_polyploid: list of Species objects
        tree_repeats: number of trees per locus
    
    Yields:
        newick strings with modified leaf labels
    """
    
    # forbidden characters within a node label: 
    # whitespace, parentheses, square brackets, comma, colon, semicolon
    non_label_chars = "([\\(\\)\\[\\],;:\\s])"  # enclosed within substitution group
    for i, line in enumerate(input_data):
        j = i//tree_repeats  # locus/marker index
        for sp in species_polyploid:
            for acc in sp.accessions:
                for allele in acc.alleles:
                    search_pattern = non_label_chars + re.escape(allele) + non_label_chars
                    replacement_string = r"\g<1>" + allele + "_m" + str(j) + r"\g<2>"
                    line = re.sub(search_pattern, replacement_string, line)
        yield line


def read_trees(fname, species_polyploid, tree_repeats):
    """
    Reads input gene trees.
    
    Args:
        fname: tree file (one newick string per line)
        species_polyploid: list of Species objects
        tree_repeats: number of trees per locus
                
    Returns:
        list of parsed trees
    """
    
    # remove trailing whitespace from each line
    with open(fname, "r") as f:
        input_data = [line.rstrip() for line in f.readlines()]

    # count and remove trailing newlines
    nt = 0
    for line in reversed(input_data):
        if line:
            break
        else:
            nt += 1
    if nt > 0:
        del input_data[-nt:]

    input_data = rename_polyploids(input_data, species_polyploid, tree_repeats) # append marker IDs
    input_data = io.StringIO("\n".join(input_data))  # create file handle
    input_data_trees = list(Phylo.parse(input_data, "newick"))

    return input_data_trees


def prune_tree(tree, alleles, suffix=""):
    """
    Prunes all leaves from a given tree which completely match one of a 
    given list of allele names. The matching pattern can be extended by 
    an optional suffix (regular expressions allowed).
    
    Args:
        tree: tree object to be pruned in place
        alleles: list of IDs
        suffix
    """
    for a in alleles:
        # search pattern matching preprocessed allele ID
        pattern = "^" + re.escape(a) + suffix + "$"
        # prune all matching leaves
        leaves_2b_pruned = [leaf.name for leaf in tree.get_terminals() if re.match(pattern, leaf.name)]
        for leaf in leaves_2b_pruned:
            tree.prune(name=leaf)


def build_mapping(prefix, allele_combinations):
    """
    Builds an allele mapping string for pseudo-diploids. For each 
    non-empty subset of alleles one hypothetical species is created.
    
    Args:
        prefix: controls name of created pseudo-diploids
        allele_combinations: list of strings, IDs already concatenated
                             for each pseudo-diploid
    
    Returns:
        allele mapping string
    """
    if len(allele_combinations) > 99:
        sys.exit("Error: maximal number of ancestors exceeded")
    
    allele_combinations = filter(None, allele_combinations)  # remove empty entries
    
    # create new species IDs incl. allele mapping
    species_list = [ "".join([ prefix, "___", "{0:02d}".format(j+1), ":", a]) 
                     for j, a in enumerate(allele_combinations) ]
    allele_map = ";".join(species_list)
    
    return allele_map


def build_PhyloNet_instruction(tree_IDs, mapping_orig, mapping_hyp, conf):
    """
    Creates an instruction for PhyloNet. Allele mappings for either true
    and hypothetical species are combined.
    
    Args:
        tree_IDs: specifies gene trees to be used
        mapping_orig: allele mapping string (diploids)
        mapping_hyp: allele mapping string (pseudo-diploids)
        conf: namespace of program options
    
    Returns:
        Instruction for species tree inference.
    """
    mapping_new = ";".join(filter(None, [mapping_orig, mapping_hyp]))
    if not mapping_new:
        sys.exit("Error: cannot build PhyloNet instruction, empty allele mapping")
        
    instruction = "".join([conf.phylonet_command, "(", tree_IDs, ") -a\n<", mapping_new, ">", ";\n"])
    return instruction


def print_results(title, tree, mapping, dist, conf):
    """
    Prints, writes, and optionally plots results.
    
    Args:
        title
        tree: inferred species tree (string)
        mapping: optimized allele mapping (string)
        dist: number of extra lineages
        conf: namespace of program options
    
    Warning:
        If the species tree is rerooted using an outgroup, it is not 
        checked wether the non-outgroup taxa are monophyletic!
    """
    
    # store newick string in place
    newick_line = io.StringIO()
    newick_line.write(tree)
    newick_line.seek(0)

    # modify tree
    final_tree = Phylo.read(newick_line, "newick")
    
    if conf.outgroup:
        final_tree.root_with_outgroup({'name':conf.outgroup[0]})
    if not conf.display_branch_length:
        clearStringIO(newick_line)
        # plain=True: branch lengths removed
        Phylo.write(final_tree, newick_line, "newick", plain=True)
        newick_line.seek(0)
        final_tree = Phylo.read(newick_line, "newick")

    outfile = open(os.path.join(conf.working_directory, "results.txt"), "a")
    result = "".join(["\n", title, ":\n", newick_line.getvalue(), "\nAllele mapping:\n", 
                      mapping, "\nTotal number of extra lineages: ", str(dist), "\n"])
    outfile.write(result)
    print(result)

    if conf.print_ascii_tree == True:
        Phylo.draw_ascii(final_tree)
        Phylo.draw_ascii(final_tree, outfile)
    
    outfile.close()


def read_allele_mapping(fname):
    """
    Reads name, species membership, allele names and ploidy level of 
    each accession.
    
    Args:
        fname: name of allele mapping file
    
    Returns:
        list of Species objects
    """
    species_list = []
    
    with open(fname, "r") as f:
        for i, line in enumerate(f, start=1):
            cols = line.split("\t")
            if len(cols) == 4:
                name = cols[0].strip()
                species = cols[1].strip()
                alleles = [ acc.strip() for acc in cols[2].split(",") if acc.strip() ]
                ploidy = int(cols[3].strip())
                
                if ploidy % 2:
                    sys.exit("Error: cannot handle uneven ploidy level")
                
                if name and species and alleles and ploidy:
                    try:
                        sp_ind = [sp.name for sp in species_list].index(species)  # species already known
                    except ValueError:
                        sp_ind = -1  # new species
                    
                    if sp_ind >= 0:
                        if ploidy != species_list[sp_ind].ploidy:
                            sys.exit("Error: conflicting ploidy levels within " + species)
                        species_list[sp_ind].accessions.append(Accession(name, alleles))  # update existing species
                    else:
                        species_list.append(Species(species, ploidy, name, alleles))  # create new species object
                    
                else:
                    sys.exit("Error: empty field in line " + str(i) + " of " + fname)
            
            elif len(cols) == 1:
                if cols[0].strip():
                    sys.exit("Error: single field in line " + str(i) + " of " + fname + " (4 expected)")
                else:
                    continue  # ignore empty lines
            else:
                sys.exit("Error: " + str(len(cols)) + " fields in line " + str(i) + " of " + fname + " (4 expected)")
    
    if species_list:
        return species_list
    else:
        sys.exit("Error: no valid entries in " + fname)




#!/usr/bin/env python3
# version: 05.12.19
# Author: Ulrich Lautenschlager
# changes by Tankred Ott:
#    put functionality into run function to be able
#    to import the script into other applications

import argparse, collections, itertools, numpy, os, random, re, scipy.stats, sys, time

solution = collections.namedtuple("solution", ["comb", "cost"])


def mean_ent(cluster_assignments):
    """
    Evaluate cost function for one solution.
    """
    nrow, ncol, nrun = cluster_assignments.shape
    csum = cluster_assignments.sum(axis=2)
    m = numpy.mean([scipy.stats.entropy(row) for row in csum])
    return m


def neighborhood_sample(solution, ssize):
    """
    Returns list of moves + list of (subsampled) neighborhood solutions
    """
    n_runs = len(solution)
    n_clusters = len(solution[0])
    
    swaps_per_run = n_clusters * (n_clusters-1) // 2
    max_size = n_runs * swaps_per_run
    
    ssize = min(ssize, max_size)
    
    combinations = list(itertools.combinations(range(n_clusters), 2))
    selection = random.sample(range(max_size), ssize)
    
    nb = [0] * ssize
    
    for i_nb, i_sel in enumerate(selection):
        i_run = i_sel // swaps_per_run
        i_comb = i_sel % swaps_per_run
        
        i1, i2 = combinations[i_comb]
        swapped = solution[i_run][:]
        swapped[i1], swapped[i2] = swapped[i2], swapped[i1]
        nb[i_nb] = solution[:i_run] + [swapped] + solution[i_run+1:]
    
    return selection, nb


def run(indfile, n_iter=400, tabu_tenure=10, sample_size=None, quiet=True):
    time_start = time.time()

    clustering = []
    runs = []  # list of clusterings

    with open(indfile, "r") as fin:
        for line in fin:
            # ignore additional info
            line = re.sub(r".*:", "", line)
            # get cluster assignments
            all = re.findall(r"[0-9]*[.]?[0-9]+", line)
            if all:
                clustering.append(all)
            else:
                if clustering:
                    runs.append(clustering)
                    clustering = []
    if clustering:
        runs.append(clustering)
        
    cluster_assignments = numpy.dstack([numpy.array(r, dtype="f8") for r in runs])
    n_rows, n_clusters, n_runs = cluster_assignments.shape
    if not quiet:
        print(cluster_assignments.shape)

    max_sample_size =  n_runs * n_clusters * (n_clusters-1) // 2
    if not sample_size:
        sample_size = max_sample_size
    if not quiet:
        print("Evaluate", sample_size, "solutions per iteration.")

    tabu_list = []
    
    # initial solution
    current_solution = [random.sample(range(n_clusters), n_clusters) for i in range(n_runs)]
    best_solution = current_solution
    cost = [0] * sample_size
    cost_best = float("inf")
    
    for iteration in range(n_iter):
        
        moves, neighborhood = neighborhood_sample(current_solution, sample_size)
        
        for j, candidate in enumerate(neighborhood):
            cluster_permuted = numpy.dstack([cluster_assignments[:,candidate[k],k] for k in range(n_runs)])
            cost[j] = mean_ent(cluster_permuted)
        
        updated = False
        
        for j in numpy.argsort(cost):
            if cost[j] < cost_best:
                i_best = iteration
                current_solution = neighborhood[j]
                best_solution = current_solution
                cost_best = cost[j]
                tabu_list = [moves[j]] + tabu_list
                tabu_list = tabu_list[:tabu_tenure]
                updated = True
                break
            else:
                if moves[j] not in tabu_list:
                    current_solution = neighborhood[j]
                    tabu_list = [moves[j]] + tabu_list
                    tabu_list = tabu_list[:tabu_tenure]
                    updated = True
                    break
        
        if not updated:
            tabu_list = tabu_list[:-1]
        
        if not quiet:
            print("best solution of iteration ", str(iteration), ": ", str(cost[j]))
            print("best solution so far: ", str(cost_best))
            #print("variance: ", str(numpy.var(cost)))
            print("------------------------")
            sys.stdout.flush()
        

    time_stop = time.time()
    if not quiet:
        print("Total time elapsed:", "%.2f" % (time_stop-time_start), "s")
        print("Best solution found in iteration", str(i_best))
    
    final_perm = numpy.array(best_solution, dtype="uint8") + 1

    cluster_permuted = numpy.dstack([cluster_assignments[:, best_solution[k], k] for k in range(n_runs)])
    q_matrix_avg = numpy.divide(cluster_permuted.sum(axis=2), n_runs)

    return final_perm, q_matrix_avg


def save(indfile, final_perm, q_matrix_avg):
    spl = os.path.splitext(indfile)

    fout = spl[0] + ".permutations"
    numpy.savetxt(fout, final_perm, fmt="%u", delimiter="\t")
    
    fout = spl[0] + ".clustering"
    #numpy.savetxt(fout, cluster_permuted.sum(axis=2), fmt="%u", delimiter="\t")
    numpy.savetxt(fout, q_matrix_avg, fmt="%1.6f", delimiter="\t")


def main():
    parser = argparse.ArgumentParser(description="Align clusters among multiple clusterings.")
    parser.add_argument("indfile", help="Membership coefficents (CLUMPP-compatible format).")
    parser.add_argument("-n", "--n-iter", default=400, type=int, help="Number of iterations.")
    parser.add_argument("-s", "--sample-size", type=int, help="Number of evaluated neighborhood solutions per iteration (default: all).")
    parser.add_argument("-t", "--tabu-tenure", default=10, type=int, help="Number of forbidden reverse moves.")
    args = parser.parse_args()

    final_perm, q_matrix_avg = run(
        args.indfile,
        n_iter=args.n_iter,
        tabu_tenure=args.tabu_tenure,
        sample_size=args.sample_size,
        quiet=False
    )
    
    print("List of permutations:")
    sys.stdout.flush()
    numpy.savetxt(sys.stdout.buffer, final_perm, fmt="%u", delimiter="\t")
    
    print("------------------------")
    
    print("Q-matrix:")
    sys.stdout.flush()
    #numpy.savetxt(sys.stdout.buffer, cluster_permuted.sum(axis=2), fmt="%u", delimiter="\t")
    numpy.savetxt(sys.stdout.buffer, q_matrix_avg, fmt="%1.6f", delimiter="\t")

    # write output files

    spl = os.path.splitext(args.indfile)

    fout = spl[0] + ".permutations"
    numpy.savetxt(fout, final_perm, fmt="%u", delimiter="\t")
    
    fout = spl[0] + ".clustering"
    #numpy.savetxt(fout, cluster_permuted.sum(axis=2), fmt="%u", delimiter="\t")
    numpy.savetxt(fout, q_matrix_avg, fmt="%1.6f", delimiter="\t")
    

if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()


setup(name = "allcopol",
    version = "0.1.1",
    description = "AllCoPol: Inferring allele co-ancestry in polyploids",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/AGOberprieler/allcopol",
    author = "Ulrich Lautenschlager",
    author_email = "ulrich.lautenschlager@ur.de",
    license = "MIT",
    packages = find_packages(),
    install_requires = [ 
        "argparse", "biopython", "configargparse", "numpy", "scipy"
    ],
    entry_points = {
        "console_scripts": [
            "allcopol=allcopol.allcopol:main",
            "align_clusters=allcopol.align_clusters:main",
            "create_indfile=allcopol.create_indfile:main",
            "relabel_trees=allcopol.relabel_trees:main",
        ],
    },
    zip_safe = False,
    python_requires = ">=3.5",
)




#!/usr/bin/env python3
"""
Convert inferred allele mappings into a format that
can be used by CLUMPP or align_clusters.py.
"""

import argparse, operator, re, sys

def main():
    parser = argparse.ArgumentParser(description="Convert inferred allele mappings into a format that can be used by CLUMPP or align_clusters.py.")
    parser.add_argument("allele_mapping_file", help="Inferred allele mappings (PhyloNet format), one line per run.")
    parser.add_argument("species", help="Species which has been splitted into pseudodiploids.")
    args = parser.parse_args()

    species = args.species

    with open(args.allele_mapping_file, "r") as infile:
        txt = infile.read()

    mapping_strings = [m.strip() for m in txt.split("\n") if m.strip()]

    p = re.compile(species + "___[^;]*")

    # get max. number of clusters
    n_clusters = max([len(p.findall(mapping_strings[1])) for m in mapping_strings])


    for i, mstring in enumerate(mapping_strings):
        mappings_pseudodipl = sorted(p.findall(mstring)) # constant suffix length required
        membership = dict()

        for j, mp in enumerate(mappings_pseudodipl):
            alleles = mp.split(":")[1]
            for allele in alleles.split(","):
                membership[allele] = j

    #    for allele in sorted(membership.keys()):
    #        membership_coefficents = ["0"] * n_clusters
    #        membership_coefficents[membership[allele]] = "1"
    #        print(" ".join([str(i), str(i), "(x) 1 :"] + membership_coefficents))

        for k, allele in enumerate(sorted(membership.keys())):
            membership_coefficents = ["0"] * n_clusters
            membership_coefficents[membership[allele]] = "1"
            print(" ".join([str(k+1), str(k+1), "(x) 1 :"] + membership_coefficents))

        print("")

if __name__ == "__main__":
    main()
